@@ -1,8 +1,12 @@
 /*
  * closefs.c --- close an ext2 filesystem
  * 
- * Copyright (C) 1993, 1994 Theodore Ts'o.  This file may be redistributed
- * under the terms of the GNU Public License.
+ * Copyright (C) 1993, 1994, 1995, 1996 Theodore Ts'o.
+ *
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Public
+ * License.
+ * %End-Header%
  */
 
 #include <stdio.h>
@@ -16,11 +20,43 @@
 
 #include <linux/ext2_fs.h>
 
-#include "ext2fs.h"
+#include "ext2fsP.h"
+
+static int test_root(int a, int b)
+{
+	if (a == 0)
+		return 1;
+	while (1) {
+		if (a == 1)
+			return 1;
+		if (a % b)
+			return 0;
+		a = a / b;
+	}
+}
+
+int ext2fs_bg_has_super(ext2_filsys fs, int group_block)
+{
+#ifdef EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER
+	struct ext2fs_sb	*s;
+
+	s = (struct ext2fs_sb *) fs->super;
+	if (!(s->s_feature_ro_compat & EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
+		return 1;
+
+	if (test_root(group_block, 3) || (test_root(group_block, 5)) ||
+	    test_root(group_block, 7))
+		return 1;
+	
+	return 0;
+#else
+	return 1;
+#endif
+}
 
 errcode_t ext2fs_flush(ext2_filsys fs)
 {
-	int		i,j;
+	int		i,j,maxgroup;
 	int		group_block;
 	errcode_t	retval;
 	char		*group_ptr;
@@ -34,7 +70,7 @@ errcode_t ext2fs_flush(ext2_filsys fs)
 	fs_state = fs->super->s_state;
 
 	fs->super->s_wtime = time(NULL);
-	if (fs->flags & EXT2_SWAP_BYTES) {
+	if (fs->flags & EXT2_FLAG_SWAP_BYTES) {
 		retval = ENOMEM;
 		if (!(super_shadow = malloc(SUPERBLOCK_SIZE)))
 			goto errout;
@@ -75,7 +111,7 @@ errcode_t ext2fs_flush(ext2_filsys fs)
 	 * we exit.)
 	 */
 	fs->super->s_state &= ~EXT2_VALID_FS;
-	if (fs->flags & EXT2_SWAP_BYTES) {
+	if (fs->flags & EXT2_FLAG_SWAP_BYTES) {
 		*super_shadow = *fs->super;
 		ext2fs_swap_super(super_shadow);
 	}
@@ -85,7 +121,12 @@ errcode_t ext2fs_flush(ext2_filsys fs)
 	 * superblocks and group descriptors.
 	 */
 	group_block = fs->super->s_first_data_block;
-	for (i = 0; i < fs->group_desc_count; i++) {
+	maxgroup = (fs->flags & EXT2_FLAG_MASTER_SB_ONLY) ? 1 :
+		fs->group_desc_count;
+	for (i = 0; i < maxgroup; i++) {
+		if (!ext2fs_bg_has_super(fs, i))
+			goto next_group;
+
 		if (i !=0 ) {
 			retval = io_channel_write_blk(fs->io, group_block,
 						      -SUPERBLOCK_SIZE,
@@ -102,6 +143,7 @@ errcode_t ext2fs_flush(ext2_filsys fs)
 				goto errout;
 			group_ptr += fs->blocksize;
 		}
+	next_group:
 		group_block += EXT2_BLOCKS_PER_GROUP(fs->super);
 	}
 
@@ -116,11 +158,10 @@ errcode_t ext2fs_flush(ext2_filsys fs)
 		if (retval)
 			goto errout;
 	}
-
 	retval = 0;
 errout:
 	fs->super->s_state = fs_state;
-	if (fs->flags & EXT2_SWAP_BYTES) {
+	if (fs->flags & EXT2_FLAG_SWAP_BYTES) {
 		if (super_shadow)
 			free(super_shadow);
 		if (group_shadow)
@@ -140,6 +181,42 @@ errcode_t ext2fs_close(ext2_filsys fs)
 		if (retval)
 			return retval;
 	}
+	if (fs->write_bitmaps) {
+		retval = fs->write_bitmaps(fs);
+		if (retval)
+			return retval;
+	}
 	ext2fs_free(fs);
 	return 0;
 }
+
+/*
+ * This procedure frees a badblocks list.
+ */
+void ext2fs_badblocks_list_free(ext2_badblocks_list bb)
+{
+	if (bb->magic != EXT2_ET_MAGIC_BADBLOCKS_LIST)
+		return;
+
+	if (bb->list)
+		free(bb->list);
+	bb->list = 0;
+	free(bb);
+}
+
+/*
+ * Close a directory block list
+ */
+void ext2fs_free_dblist(ext2_dblist dblist)
+{
+	if (!dblist || (dblist->magic != EXT2_ET_MAGIC_DBLIST))
+		return;
+
+	if (dblist->list)
+		free(dblist->list);
+	dblist->list = 0;
+	if (dblist->fs && dblist->fs->dblist == dblist)
+		dblist->fs->dblist = 0;
+	dblist->magic = 0;
+	free(dblist);
+}
