@@ -1,8 +1,12 @@
 /*
  * mke2fs.c - Make a ext2fs filesystem.
  * 
- * Copyright (C) 1994 Theodore Ts'o.  This file may be redistributed
- * under the terms of the GNU Public License.
+ * Copyright (C) 1994, 1995, 1996, 1997 Theodore Ts'o.
+ *
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Public
+ * License.
+ * %End-Header%
  */
 
 /* Usage: mke2fs [options] device
@@ -47,11 +51,16 @@
 #endif
 
 #include "et/com_err.h"
+#include "uuid/uuid.h"
 #include "ext2fs/ext2fs.h"
 #include "../version.h"
 
 #define STRIDE_LENGTH 8
 
+#ifndef sparc
+#define ZAP_BOOTBLOCK
+#endif
+
 extern int isatty(int);
 extern FILE *fpopen(const char *cmd, const char *mode);
 
@@ -67,14 +76,17 @@ int	force = 0;
 char	*bad_blocks_filename = 0;
 
 struct ext2_super_block param;
+char *creator_os = NULL;
+char *volume_label = NULL;
+char *mount_dir = NULL;
 
 static void usage(NOARGS)
 {
-	fprintf(stderr,
-		"Usage: %s [-c|-t|-l filename] [-b block-size] "
-		"[-f fragment-size]\n\t[-i bytes-per-inode] "
-		"[-m reserved-blocks-percentage] [-qvS]\n"
-		"\t[-g blocks-per-group] device [blocks-count]\n",
+	fprintf(stderr, "Usage: %s [-c|-t|-l filename] [-b block-size] "
+	"[-f fragment-size]\n\t[-i bytes-per-inode] "
+	"[-m reserved-blocks-percentage] [-qvS]\n\t"
+	"[-o creator-os] [-g blocks-per-group] [-L volume-label]\n\t"
+	"[-M last-mounted-directory] device [blocks-count]\n",
 		program_name);
 	exit(1);
 }
@@ -100,7 +112,11 @@ static void check_plausibility(NOARGS)
 	val = stat(device_name, &s);
 	
 	if(val == -1) {
-		perror("stat");
+		printf("Could not stat %s --- %s\n", device_name,
+		       error_message(errno));
+		if (errno == ENOENT)
+			printf("\nThe device apparently does not exist; "
+			       "did yo specify it correctly?\n");
 		exit(1);
 	}
 	if(!S_ISBLK(s.st_mode)) {
@@ -277,49 +293,12 @@ static void handle_bad_blocks(ext2_filsys fs, badblocks_list bb_list)
 	badblocks_list_iterate_end(bb_iter);
 }
 
-static void new_table_block(ext2_filsys fs, blk_t first_block,
-			    const char *name, int num, int initialize,
-			    const char *buf, blk_t *new_block)
+static void write_inode_tables(ext2_filsys fs)
 {
 	errcode_t	retval;
 	blk_t		blk;
-	int		i;
-	int		count;
-	
-	retval = ext2fs_get_free_blocks(fs, first_block,
-			first_block + fs->super->s_blocks_per_group,
-					num, fs->block_map, new_block);
-	if (retval) {
-		printf("Could not allocate %d block(s) for %s: %s\n",
-		       num, name, error_message(retval));
-		ext2fs_unmark_valid(fs);
-		return;
-	}
-	if (initialize) {
-		blk = *new_block;
-		for (i=0; i < num; i += STRIDE_LENGTH, blk += STRIDE_LENGTH) {
-			if (num-i > STRIDE_LENGTH)
-				count = STRIDE_LENGTH;
-			else
-				count = num - i;
-			retval = io_channel_write_blk(fs->io, blk, count, buf);
-			if (retval)
-				printf("Warning: could not write %d blocks "
-				       "starting at %d for %s: %s\n",
-				       count, blk, name,
-				       error_message(retval));
-		}
-	}
-	blk = *new_block;
-	for (i = 0; i < num; i++, blk++)
-		ext2fs_mark_block_bitmap(fs->block_map, blk);
-}	
-
-static void alloc_tables(ext2_filsys fs)
-{
-	blk_t	group_blk;
-	int	i;
-	char	*buf;
+	int		i, j, num, count;
+	char		*buf;
 
 	buf = malloc(fs->blocksize * STRIDE_LENGTH);
 	if (!buf) {
@@ -328,25 +307,30 @@ static void alloc_tables(ext2_filsys fs)
 	}
 	memset(buf, 0, fs->blocksize * STRIDE_LENGTH);
 	
-	group_blk = fs->super->s_first_data_block;
 	if (!quiet)
 		printf("Writing inode tables: ");
 	for (i = 0; i < fs->group_desc_count; i++) {
 		if (!quiet)
 			printf("%4d/%4ld", i, fs->group_desc_count);
-		new_table_block(fs, group_blk, "block bitmap", 1, 0, buf,
-				&fs->group_desc[i].bg_block_bitmap);
-		new_table_block(fs, group_blk, "inode bitmap", 1, 0, buf,
-				&fs->group_desc[i].bg_inode_bitmap);
-		new_table_block(fs, group_blk, "inode table",
-				fs->inode_blocks_per_group,
-				!super_only, buf,
-				&fs->group_desc[i].bg_inode_table);
 		
-		group_blk += fs->super->s_blocks_per_group;
+		blk = fs->group_desc[i].bg_inode_table;
+		num = fs->inode_blocks_per_group;
+		
+		for (j=0; j < num; j += STRIDE_LENGTH, blk += STRIDE_LENGTH) {
+			if (num-j > STRIDE_LENGTH)
+				count = STRIDE_LENGTH;
+			else
+				count = num - j;
+			retval = io_channel_write_blk(fs->io, blk, count, buf);
+			if (retval)
+				printf("Warning: could not write %d blocks "
+				       "in inode table starting at %d: %s\n",
+				       count, blk, error_message(retval));
+		}
 		if (!quiet) 
 			printf("\b\b\b\b\b\b\b\b\b");
 	}
+	free(buf);
 	if (!quiet)
 		printf("done     \n");
 }
@@ -368,7 +352,9 @@ static void create_root_dir(ext2_filsys fs)
 				"while reading root inode");
 			exit(1);
 		}
-		inode.i_uid = geteuid();
+		inode.i_uid = getuid();
+		if (inode.i_uid)
+			inode.i_gid = getgid();
 		retval = ext2fs_write_inode(fs, EXT2_ROOT_INO, &inode);
 		if (retval) {
 			com_err("ext2fs_write_inode", retval,
@@ -437,6 +423,7 @@ static void reserve_inodes(ext2_filsys fs)
 	ext2fs_mark_ib_dirty(fs);
 }
 
+#ifdef ZAP_BOOTBLOCK
 static void zap_bootblock(ext2_filsys fs)
 {
 	char buf[512];
@@ -449,11 +436,13 @@ static void zap_bootblock(ext2_filsys fs)
 		printf("Warning: could not erase block 0: %s\n", 
 		       error_message(retval));
 }
+#endif
 	
 
 static void show_stats(ext2_filsys fs)
 {
-	struct ext2_super_block	*s = fs->super;
+	struct ext2fs_sb 	*s = (struct ext2fs_sb *) fs->super;
+	char 			buf[80];
 	blk_t			group_block;
 	int			i, col_left;
 	
@@ -461,6 +450,16 @@ static void show_stats(ext2_filsys fs)
 		printf("warning: %d blocks unused.\n\n",
 		       param.s_blocks_count - s->s_blocks_count);
 	
+	switch (fs->super->s_creator_os) {
+	    case EXT2_OS_LINUX: printf ("Linux"); break;
+	    case EXT2_OS_HURD:  printf ("GNU/hurd");   break;
+	    case EXT2_OS_MASIX: printf ("Masix"); break;
+	    default:		printf ("(unknown os)");
+        }
+	printf (" ext2 filesystem format\n");
+	memset(buf, 0, sizeof(buf));
+	strncpy(buf, s->s_volume_name, sizeof(s->s_volume_name));
+	printf("Filesystem label=%s\n", buf);
 	printf("%u inodes, %u blocks\n", s->s_inodes_count,
 	       s->s_blocks_count);
 	printf("%u blocks (%2.2f%%) reserved for the super user\n",
@@ -487,6 +486,8 @@ static void show_stats(ext2_filsys fs)
 	col_left = 0;
 	for (i = 1; i < fs->group_desc_count; i++) {
 		group_block += s->s_blocks_per_group;
+		if (!ext2fs_bg_has_super(fs, i))
+			continue;
 		if (!col_left--) {
 			printf("\n\t");
 			col_left = 8;
@@ -498,6 +499,41 @@ static void show_stats(ext2_filsys fs)
 	printf("\n\n");
 }
 
+#ifndef HAVE_STRCASECMP
+static int strcasecmp (char *s1, char *s2)
+{
+	while (*s1 && *s2) {
+		int ch1 = *s1++, ch2 = *s2++;
+		if (isupper (ch1))
+			ch1 = tolower (ch1);
+		if (isupper (ch2))
+			ch2 = tolower (ch2);
+		if (ch1 != ch2)
+			return ch1 - ch2;
+	}
+	return *s1 ? 1 : *s2 ? -1 : 0;
+}
+#endif
+
+/*
+ * Set the S_CREATOR_OS field.  Return true if OS is known,
+ * otherwise, 0.
+ */
+static int set_os(struct ext2_super_block *sb, char *os)
+{
+	if (isdigit (*os))
+		sb->s_creator_os = atoi (os);
+	else if (strcasecmp(os, "linux") == 0)
+		sb->s_creator_os = EXT2_OS_LINUX;
+	else if (strcasecmp(os, "GNU") == 0 || strcasecmp(os, "hurd") == 0)
+		sb->s_creator_os = EXT2_OS_HURD;
+	else if (strcasecmp(os, "masix") == 0)
+		sb->s_creator_os = EXT2_OS_MASIX;
+	else
+		return 0;
+	return 1;
+}
+
 #define PATH_SET "PATH=/sbin"
 
 static void PRS(int argc, char *argv[])
@@ -505,11 +541,14 @@ static void PRS(int argc, char *argv[])
 	char	c;
 	int	size;
 	char	* tmp;
+	blk_t	max = 8192;
 	int	inode_ratio = 4096;
 	int	reserved_ratio = 5;
 	errcode_t	retval;
+	int	sparse_option = -1;
 	char	*oldpath = getenv("PATH");
-
+	struct ext2fs_sb *param_ext2 = (struct ext2fs_sb *) &param;
+	
 	/* Update our PATH to include /sbin  */
 	if (oldpath) {
 		char *newpath;
@@ -526,13 +565,17 @@ static void PRS(int argc, char *argv[])
 	setbuf(stderr, NULL);
 	initialize_ext2_error_table();
 	memset(&param, 0, sizeof(struct ext2_super_block));
+#ifdef EXT2_DYNAMIC_REV
+	param.s_rev_level = EXT2_DYNAMIC_REV;
+#endif
 	
 	fprintf (stderr, "mke2fs %s, %s for EXT2 FS %s, %s\n",
 		 E2FSPROGS_VERSION, E2FSPROGS_DATE,
 		 EXT2FS_VERSION, EXT2FS_DATE);
 	if (argc && *argv)
 		program_name = *argv;
-	while ((c = getopt (argc, argv, "b:cf:g:i:l:m:qr:tvI:SF")) != EOF)
+	while ((c = getopt (argc, argv,
+			    "b:cf:g:i:l:m:o:qr:s:tvI:SFL:M:")) != EOF)
 		switch (c) {
 		case 'b':
 			size = strtoul(optarg, &tmp, 0);
@@ -543,6 +586,7 @@ static void PRS(int argc, char *argv[])
 			}
 			param.s_log_block_size =
 				log2(size >> EXT2_MIN_BLOCK_LOG_SIZE);
+			max = size * 8;
 			break;
 		case 'c':
 		case 't':	/* Check for bad blocks */
@@ -567,12 +611,6 @@ static void PRS(int argc, char *argv[])
 					"Illegal number for blocks per group");
 				exit(1);
 			}
-			if (param.s_blocks_per_group < 256 ||
-			    param.s_blocks_per_group > 8192 || *tmp) {
-				com_err(program_name, 0,
-					"blocks per group count out of range");
-				exit(1);
-			}
 			if ((param.s_blocks_per_group % 8) != 0) {
 				com_err(program_name, 0,
 				"blocks per group must be multiple of 8");
@@ -606,9 +644,15 @@ static void PRS(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case 'o':
+			creator_os = optarg;
+			break;
 		case 'r':
 			param.s_rev_level = atoi(optarg);
 			break;
+		case 's':
+			sparse_option = atoi(optarg);
+			break;
 #ifdef EXT2_DYNAMIC_REV
 		case 'I':
 			param.s_inode_size = atoi(optarg);
@@ -623,6 +667,12 @@ static void PRS(int argc, char *argv[])
 		case 'F':
 			force = 1;
 			break;
+		case 'L':
+			volume_label = optarg;
+			break;
+		case 'M':
+			mount_dir = optarg;
+			break;
 		case 'S':
 			super_only = 1;
 			break;
@@ -655,12 +705,21 @@ static void PRS(int argc, char *argv[])
 						EXT2_BLOCK_SIZE(&param),
 						&param.s_blocks_count);
 		if (retval) {
-			com_err(program_name, 0,
+			com_err(program_name, retval,
 				"while trying to determine filesystem size");
 			exit(1);
 		}
 	}
 
+	if (param.s_blocks_per_group) {
+		if (param.s_blocks_per_group < 256 ||
+		    param.s_blocks_per_group > max || *tmp) {
+			com_err(program_name, 0,
+				"blocks per group count out of range");
+			exit(1);
+		}
+	}
+
 	/*
 	 * Calculate number of inodes based on the inode ratio
 	 */
@@ -672,6 +731,20 @@ static void PRS(int argc, char *argv[])
 	 * Calculate number of blocks to reserve
 	 */
 	param.s_r_blocks_count = (param.s_blocks_count * reserved_ratio) / 100;
+
+	/*
+	 * If we are using revision #1, use the sparse super feature
+	 * by default
+	 */
+#ifdef EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER
+	if ((sparse_option == 1)
+#ifdef EXT2_DYNAMIC_REV
+	    || (param.s_rev_level >= EXT2_DYNAMIC_REV) && (!sparse_option)
+#endif
+	    ) 
+		param_ext2->s_feature_ro_compat |=
+			EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
+#endif
 }
 					
 int main (int argc, char *argv[])
@@ -679,6 +752,7 @@ int main (int argc, char *argv[])
 	errcode_t	retval = 0;
 	ext2_filsys	fs;
 	badblocks_list	bb_list = 0;
+	struct ext2fs_sb *s;
 	
 	PRS(argc, argv);
 
@@ -692,6 +766,38 @@ int main (int argc, char *argv[])
 		exit(1);
 	}
 
+	/*
+	 * Generate a UUID for it...
+	 */
+	s = (struct ext2fs_sb *) fs->super;
+	uuid_generate(s->s_uuid);
+
+	/*
+	 * Override the creator OS, if applicable
+	 */
+	if (creator_os && !set_os(fs->super, creator_os)) {
+		com_err (program_name, 0, "unknown os - %s", creator_os);
+		exit(1);
+	}
+
+	/*
+	 * Set the volume label...
+	 */
+	if (volume_label) {
+		memset(s->s_volume_name, 0, sizeof(s->s_volume_name));
+		strncpy(s->s_volume_name, volume_label,
+			sizeof(s->s_volume_name));
+	}
+
+	/*
+	 * Set the last mount directory
+	 */
+	if (mount_dir) {
+		memset(s->s_last_mounted, 0, sizeof(s->s_last_mounted));
+		strncpy(s->s_last_mounted, mount_dir,
+			sizeof(s->s_last_mounted));
+	}
+	
 	if (!quiet)
 		show_stats(fs);
 
@@ -701,16 +807,24 @@ int main (int argc, char *argv[])
 		test_disk(fs, &bb_list);
 
 	handle_bad_blocks(fs, bb_list);
-	alloc_tables(fs);
+	retval = ext2fs_allocate_tables(fs);
+	if (retval) {
+		com_err(program_name, retval,
+			"while trying to allocate filesystem tables");
+		exit(1);
+	}
 	if (super_only) {
 		fs->super->s_state |= EXT2_ERROR_FS;
 		fs->flags &= ~(EXT2_FLAG_IB_DIRTY|EXT2_FLAG_BB_DIRTY);
 	} else {
+		write_inode_tables(fs);
 		create_root_dir(fs);
 		create_lost_and_found(fs);
 		reserve_inodes(fs);
 		create_bad_block_inode(fs, bb_list);
+#ifdef ZAP_BOOTBLOCK
 		zap_bootblock(fs);
+#endif
 	}
 	
 	if (!quiet)
